// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_CURVED_SURFACE_GRID_INDEXSETS_HH
#define DUNE_CURVED_SURFACE_GRID_INDEXSETS_HH

#include <type_traits>
#include <vector>

#include <dune/grid/common/gridenums.hh>
#include <dune/grid/common/indexidset.hh>

namespace Dune
{
  namespace Curved
  {

    // IndexSet
    // --------

    template< class Grid, class HostIndexSet >
    class IndexSet
        : public Dune::IndexSet<Grid, IndexSet<Grid, HostIndexSet>, typename HostIndexSet::IndexType, typename HostIndexSet::Types>
    {
      using Self = IndexSet;
      using Super = Dune::IndexSet<Grid, Self, typename HostIndexSet::IndexType, typename HostIndexSet::Types>;

      using Traits = typename std::remove_const_t<Grid>::Traits;
      using HostGrid = typename Traits::HostGrid;

    public:
      static const int dimension = Traits::dimension;

      using IndexType = typename Super::IndexType;

      using Types = typename Super::Types;

    public:
      IndexSet () = default;

      explicit IndexSet (const HostIndexSet& hostIndexSet)
        : hostIndexSet_(&hostIndexSet)
      {}

      template< int cc >
      IndexType index (const typename Traits::template Codim<cc>::Entity& entity) const
      {
        return hostIndexSet().index(Grid::template getHostEntity<cc>(entity));
      }

      using Super::index;

      template< int cc >
      IndexType subIndex (const typename Traits::template Codim<cc>::Entity& entity, int i, unsigned int codim) const
      {
        return hostIndexSet().subIndex(Grid::template getHostEntity<cc>(entity), i, codim);
      }

      using Super::subIndex;

      IndexType size (GeometryType type) const
      {
        return hostIndexSet().size(type);
      }

      int size (int codim) const
      {
        return hostIndexSet().size(codim);
      }

      template< class Entity >
      bool contains (const Entity& entity) const
      {
        return hostIndexSet().contains(Grid::template getHostEntity<Entity::codimension>(entity));
      }

      Types types (int codim) const
      {
        return hostIndexSet().types(codim);
      }

      const std::vector<GeometryType>& geomTypes (int codim) const
      {
        return hostIndexSet().geomTypes(codim);
      }

      explicit operator bool () const { return bool(hostIndexSet_); }

      void reset () { hostIndexSet_ = nullptr; }
      void reset (const HostIndexSet& hostIndexSet) { hostIndexSet_ = &hostIndexSet; }

    private:
      const HostIndexSet& hostIndexSet () const
      {
        assert( *this );
        return *hostIndexSet_;
      }

      const HostIndexSet* hostIndexSet_ = nullptr;
    };

  } // namespace Curved
} // namespace Dune

#endif // DUNE_CURVED_SURFACE_GRID_INDEXSETS_HH
